/** 
 *   FIXME ....
 *
 *   Commands definitions for "antenna" position system.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ??? 
 **/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <malloc.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/errno.h>

#include "command.h"
#include "iar_bus.h"
#include "iar.h"
#include "my_stdio.h"
#include "survey.h"
#include "iar_bus.h"
#include "antena.h"
#include "coordinate.h"
#include "encoders.h"
#include "anemometer.h"

/* A structure which contains information on the commands this program
   can understand. */

typedef struct cm {
	char *name;		/* User printable name of the function. */
	/* Function to call to do the job. */
	int (*func)(int fd_out, char *name);
	char *doc;		/* Documentation for this function.  */
	struct cm *extended;
}COMMAND;

//typedef struct com COMMAND;

/* The names of functions that actually do the manipulation. */
int com_help(int fd_out, char *name);
int com_quit(int fd_out, char *name);
int com_getraw(int fd_out, char *name);
int com_toggle_echo(int fd_out, char *name);

int com_start(int fd_out, char * name);
int com_stop(int fd_out, char * name);
int com_status(int fd_out, char * name);
int com_show(int fd_out, char * name);
int com_hide(int fd_out, char * name);
int com_set(int fd_out, char * name);
int com_set_zero(int fd_out, char * name);
int com_set_cw(int fd_out, char * name);
int com_set_ccw(int fd_out, char * name);
int com_test_sclk(int fd_out, char * name);
int com_put_fail_clk(int fd_out, char * name);
int com_put_normal_clk(int fd_out, char * name);

COMMAND commands_set[] = {
	{ "zero", com_set_zero, 
	  "Set zero for current position on encoder \"dec\" or \"ra\"" },
	{ "cw", com_set_cw, 
	  "Set clockwhise increment on encoder \"dec\" or \"ra\"" },
	{ "ccw", com_set_ccw, 
	  "Set counter clockwhise increment on encoder \"dec\" or \"ra\"" },
	{ (char *)NULL, (int) NULL, (char *)NULL }
};

COMMAND commands_main[] = {
	{ "set", com_set, 
	  "Set some encoders attributes (see help)", commands_set},
	{ "start", com_start, "Start acquisition", NULL},
	{ "stop", com_stop, "Stop acquisition", NULL },
	{ "status", com_status, "Status of survey file", NULL },
	{ "show", com_show, "Show positon continuously"},
	{ "hide", com_hide, "Not show positon continuously"},
	{ "getenc", com_getraw, "Get raw position of encoders" },
	{ "testsclk", com_test_sclk, "Test if SCLK is running."},
	{ "failclk", com_put_fail_clk, 
	  "Put system in fail condition when SCLK not running"},
	{ "normalclk", com_put_normal_clk,
	  "Put system in normal condition when SCLK running again"},
	{ "e", com_toggle_echo, "Toggle remote echo (for Win telnet clients)"},
	{ "help", com_help, "Display this text"},
	{ "?", com_help, "Synonym for `help'"},
	{ "quit", com_quit, "Quit ..."},
//	{ "q", com_quit, "Quit ..."},
	{ (char *)NULL, (int) NULL, (char *)NULL}
};


/* When non-zero, this means the user is done using this program. */
static int done = 0;
static int remote_echo = 0;
static int track_pos = 0;
static int fd_track = -1;
static int ra, dec;

/* Look up NAME as the name of a command, and return a pointer to that
   command.  Return a NULL pointer if NAME isn't a command name. */
COMMAND *find_command (COMMAND *cm, char *name)
{
	int i;

	for (i = 0; cm[i].name; i++)
		if (strcmp (name, cm[i].name) == 0)
			return (&cm[i]);
	
	return ((COMMAND *)NULL);
}


/* Execute a command line. */
int execute_line (int fd_out, COMMAND *cm, char *line)
{
	int i;
	COMMAND *command;
	char *word;
	
	/* Isolate the command word. */
	i = 0;
	while (line[i] && whitespace (line[i]))	i++;
	word = line + i;
	
	while (line[i] && !whitespace (line[i])) i++;

	if (line[i])
		line[i++] = '\0';

	command = find_command (cm, word);

	if (!command) {
		fd_printf (fd_out, 
			   "# %s: No such command for ""FIXME"".\r\n", word);
		return (-1);
	}

	/* Get argument to command, if any. */
	while (whitespace (line[i])) i++;

	word = line + i;

	/* Call the function. */
	return ((*(command->func)) (fd_out, word));
}



/* The user wishes to quit using this program.  Just set DONE
   non-zero. */
int com_quit (int fd_out, char *arg)
{
	done = 1;
	return (0);
}


int com_toggle_echo (int fd_out, char *arg)
{
	remote_echo ^= 1;
	return (0);
}


/* Print out help for ARG, or for all of the commands if ARG is not
   present. */

int com_help_real (int fd_out, char *arg, COMMAND *cm)
{
	int i, j;
	int printed = 0;
	int cmp;
	char *sub_arg = "";
	COMMAND *commands = cm;
	
	for (i = 0; commands[i].name; i++) {
		cmp = strcmp (arg, commands[i].name);
		/* FIXME: set se parece a otros comando y cmp retoran valores positivos */ 
		//printf("cmp: %d\n", cmp);
		if (!*arg || ( cmp == 0)) {
			fd_printf (fd_out, "# %-13s%s.\r\n", commands[i].name, 
				commands[i].doc);
			printed++;
		}
		/* FIXME cont: originalmente (cmp >= 0) */
		if ( commands[i].extended != NULL && (cmp == 0)){
			j = -1;
			if (cmp > 0){
				j = strlen(commands[i].name) + 1;
				while (whitespace (arg[j])) j++;
				sub_arg = arg + j;
			}
			if ( (j + 1) < strlen(arg) ){
				if ( sub_arg == "" )
					fd_printf (fd_out,
						   "# Possible %s options ar:\r\n", arg);
				com_help_real(fd_out, sub_arg, commands->extended);
				return 0;
			}
		}
	}

	if (!printed) {
		fd_printf (fd_out, 
			   "# No commands match `%s'. Possibilties are:\r\n", 
			   arg);

		fd_printf (fd_out, "# ");
		for (i = 0; commands[i].name; i++) {
			/* Print in six columns. */
			if (printed == 6) {
				printed = 0;
				fd_printf (fd_out, "\r\n# ");
			}
			
			fd_printf (fd_out, "%-13s", commands[i].name);
			printed++;
		}

		if (printed)
			fd_printf (fd_out, "\r\n");
	}
	return (0);
}


/* help front end */

int com_help (int fd_out, char *arg)
{
	return com_help_real(fd_out, arg, commands_main);
}


int manager_track(int fd_out, int fd)
{
	struct ib_packet pk;
	struct antena_packet *ap;
	struct encoders_data *d;

	ap = (struct antena_packet *)&pk.data;
	d = (struct encoders_data *)ap->data;
	ibus_read(fd, &pk);

	switch (ap->h.cmd) {
	case POSITION_ACM:
		if( track_pos ) {
			fd_printf( fd_out, 
				   "%8.4f %8.4f %3d [0x%04x 0x%04x]\r\n", 
				   ENC2DEG(d->val[ra]), 
				   ENC2DEC(d->val[dec]),
				   anemometer_conv(d->anemometer),
				   d->val[ra],
				   d->val[dec]
				);
		}
		break;
	case INFO_ACM:
		fd_printf(fd_out, "%s\r\n", ap->data);
		break;
	}

	return 0;
}



int command_main(int fd_in, int fd_out, int fd)
{
	char *line = NULL,
		*s;
	fd_set rfds;
	struct timeval tv;
	int retval;
	char c = 0;
        int l;
	
	ra  = encoder_num[RA_CASE];
	dec = encoder_num[DEC_CASE];
	
	fd_track = fd;
	ibus_on( fd );
	done = 0;
	line = (char *) malloc(COMMAND_LINE_LEGEND);

	/* Loop reading and executing lines until the user quits. */
	while( !done ) {
		l = 0;
		do {
			FD_ZERO(&rfds);
			FD_SET(fd_in, &rfds);
			FD_SET(fd, &rfds);
			tv.tv_sec = 10;
			tv.tv_usec = 0;
			retval = select(50, &rfds, NULL, NULL, &tv);
			c = 0;
			if (retval) {
				if (FD_ISSET(fd_in, &rfds)) {
				/* Win telnet clients send char by char ... */
					read(fd_in, &c, 1); 
					*(line+l) = c;
					l++;
					if ( remote_echo ) 
						fd_printf(fd_out,"%c", c);
				}

				if (FD_ISSET(fd, &rfds)) {
					manager_track(fd_out, fd);
				}
			}
		} while(c != '\n' && l < COMMAND_LINE_LEGEND);
		if ( *(line+l-2) == 0xd ) *(line+l-2) = 0;
		else *(line+l-1) = 0;

		if (!line)
			break;
		/* Remove leading and trailing whitespace from the
		   line.  Then, if there is anything left, add it to
		   the history list and execute it. */
		s = stripwhite (line);
		if (*s){
			execute_line (fd_out, commands_main, s);
		}
		
	}
	ibus_remove( fd );		
	free (line);
	/* fd_printf (fd_out, "# connection closed\r\n"); */

	return 0;
}


int com_set(int fd_out, char * name)
{
	if ( *name == '\0') {
		fd_printf (fd_out, "# set: missing argument.\r\n");
		return (-1);
	}

	return execute_line(fd_out, commands_set, name);
}


int com_start(int fd_out, char * name)
{
	struct antena_packet_head aph;
	
	aph.cmd = SURVEY_START_ACM;
	return ibus_send_data( fd_track, &aph, 
		       sizeof(struct antena_packet_head));
}


int com_stop(int fd_out, char * name)
{
	struct antena_packet_head aph;
	
	aph.cmd = SURVEY_STOP_ACM;
	return ibus_send_data( fd_track, &aph, 
			       sizeof(struct antena_packet_head));
}

int com_status(int fd_out, char * name)
{
	struct antena_packet_head aph;
	
	aph.cmd = SURVEY_STATUS_ACM;
	return ibus_send_data( fd_track, &aph, 
			       sizeof(struct antena_packet_head));
}


int com_show(int fd_out, char * name)
{
	track_pos = 1;

	return 0;
}


int com_hide(int fd_out, char * name)
{
	track_pos = 0;

	return 0;
}




